#!/bin/bash
#
# Anbox - The Android in a Box runtime environment
# Copyright 2020 Canonical Ltd.  All rights reserved.
#

## check_service.sh calls the /status endpoint of the Stream Gateway service and
## returns a Nagios status code based on the result

EXPECTED_DB_NODES="unknown"

# Nagios return codes
OK=0
WARNING=1
CRITICAL=2
UNKNOWN=3

EXIT_CODE="$OK"

for p in "$@"
do
  case $p in
    --expected-db-nodes=*)
        EXPECTED_DB_NODES=${p#*=}
        ;;
    *)
      echo "INVALID CHECK - unrecognized option $p"
      exit $UNKNOWN
  esac
done

# set_return_code sets the check exit code only if it's more severe than the current one
# e.g. you cannot go from CRITICAL to OK
function set_return_code() {
  local code=$1
  if  [ -z $code ] || [ "$code" -le "$EXIT_CODE" ]; then
    return
  fi
  EXIT_CODE="$code"
}

status_output=$(curl --unix-socket /var/snap/anbox-stream-gateway/common/service/internal.socket -X GET http:/internal/1.0/status)
if [ $? -ne 0 ]; then
  echo "GATEWAY CRITICAL - Failed to get status"
  echo "$status_output"
  exit $CRITICAL
fi

status=$(echo "$status_output" | jq '.metadata.status')
agents=$(echo "$status_output" | jq '.metadata.agents')
db_nodes=$(echo "$status_output" | jq '.metadata.database_nodes')
if [ "$status" = "unhealthy" ]; then
  set_return_code $CRITICAL
fi
if [ "$EXPECTED_DB_NODES" != "unknown" ] && [ "$db_nodes" -ne "$EXPECTED_DB_NODES" ]; then
  set_return_code $WARNING
fi

# Print correct status and debug information
case $EXIT_CODE in
    $OK)
        echo "GATEWAY OK"
        ;;
    $WARNING)
        echo "GATEWAY WARNING"
        ;;
    $CRITICAL)
        echo "GATEWAY CRITICAL"
        ;;
    $UNKNOWN)
        echo "GATEWAY UNKNOWN"
        ;;
  esac

echo "Agents: $agents"
if [ "$EXPECTED_DB_NODES" != "unknown" ]; then
  echo "Database nodes: $db_nodes - Expected $EXPECTED_DB_NODES"
else
  echo "Database nodes: $db_nodes"
fi

exit $EXIT_CODE
