#
# Anbox - The Android in a Box runtime environment
# Copyright 2019 Canonical Ltd.  All rights reserved.
#

import json
import os
import random
import shutil
import socket
import string
import subprocess
from subprocess import (check_call, check_output,
                        CalledProcessError, STDOUT)

import requests_unixsocket
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import templating
from charmhelpers.core.hookenv import (log, unit_public_ip, charm_name,
                                       status_set, unit_private_ip,
                                       config, open_port,
                                       local_unit, relation_id, relation_set,
                                       remote_unit, is_leader, leader_get,
                                       leader_set, relation_ids, application_version_set)
from charmhelpers.core.host import (add_group, service, service_restart, add_user_to_group,
                                    service_stop)
from charmhelpers.core.unitdata import kv
from charms.layer.anbox import (get_public_address,
                                reload_systemd,
                                close_old_and_open_new_ports,
                                generate_random_string,
                                install_or_upgrade_anbox_package,
                                remove_anbox_package)
from charms.reactive import (hook, when, set_flag, clear_flag, is_flag_set,
                             when_not, when_any, when_not_all)
from charms.reactive.relations import endpoint_from_flag
from yaml import dump, Dumper

SERVICE = 'snap.anbox-stream-gateway.daemon.service'
GATEWAY = 'anbox-stream-gateway'
GATEWAY_GROUP = 'anbox-stream-gateway'

SNAP_NAME = 'anbox-stream-gateway'
SNAP_COMMON_PATH = '/var/snap/anbox-stream-gateway/common'
SERVICE_PATH = '{}/service'.format(SNAP_COMMON_PATH)
GATEWAY_CONFIG_PATH = '{}/config.yaml'.format(SERVICE_PATH)
AUTH_SECRET_PATH = '{}/auth_secret'.format(SERVICE_PATH)

TLS_DIR = '{}/tls'.format(SNAP_COMMON_PATH)
TLS_PUBLIC_CERT = '{}/public_cert.pem'.format(TLS_DIR)
TLS_PUBLIC_KEY = '{}/public_key.pem'.format(TLS_DIR)
TLS_PRIVATE_CERT = '{}/private_cert.pem'.format(TLS_DIR)
TLS_PRIVATE_KEY = '{}/private_key.pem'.format(TLS_DIR)
TLS_CA_CERT = '{}/ca_cert.pem'.format(TLS_DIR)

SNAP_DQLITE_PATH = '{}/dqlite'.format(SNAP_COMMON_PATH)
DQLITE_DATA_DIR = "{}/meta".format(SNAP_DQLITE_PATH)
DQLITE_PORT_BASE = 7000

DEB_PACKAGE_NAME = "anbox-cloud-stream-gateway"

NRPE_PLUGIN_PATH = '/usr/lib/nagios/plugins'
NRPE_USER = 'nagios'

SERVICE_DROP_IN_FOLDER = '/etc/systemd/system/{}.d'.format(SERVICE)
SERVICE_DROP_IN_NAME = '10-gateway-unix-socket-chown.conf'
SERVICE_DROP_IN_PATH = '{}/{}'.format(SERVICE_DROP_IN_FOLDER,
                                      SERVICE_DROP_IN_NAME)


"""
Flags:
 - anbox-stream-gateway.installed: the stream gateway snap has been installed
 - gateway.primed: the required components and relations for the gateway to start are available
 - gateway.started: the gateway has been configured and was started
 - gateway.changed: something requires a re-configuration of the stream gateway

 - client.token_created: set when a new client connects to the interface
"""


def ensure_service_dir():
    if not os.path.exists(SERVICE_PATH):
        os.mkdir(SERVICE_PATH, mode=0o750)
    elif not (os.stat(SERVICE_PATH).st_mode & 0o777) == 0o750:
        os.chmod(SERVICE_PATH, 0o750)


@when('ua.ready')
@when_not('anbox-stream-gateway.installed')
def install():
    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
    set_charm_version()
    add_group(charm_name())
    ensure_service_dir()
    deploy_drop_in()
    configure()
    set_flag('anbox-stream-gateway.installed')


def set_charm_version():
    try:
        info = subprocess.run(['snap', 'info', 'anbox-stream-gateway'],
                              stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as err:
        log('Failed to get snap info: {}'.format(err), level='ERROR')
    else:
        stdout = info.stdout.decode('utf8').splitlines()
        version = [line.split()[1] for line in stdout if 'installed' in line]
        if version:
            application_version_set(version[0])


def is_departing_ha_relation():
    if 'JUJU_DEPARTING_UNIT' not in os.environ:
        return False

    departing_unit = os.environ['JUJU_DEPARTING_UNIT']
    current_unit = os.environ['JUJU_UNIT_NAME']
    return departing_unit == current_unit


@when('endpoint.ha.departed', 'gateway.started')
def ha_relation_departed():
    if not is_departing_ha_relation():
        log('We are not the departing node')
        return

    address = kv().get('dqlite-address')
    if len(address) == 0:
        raise Exception('Gateway was running without a configured dqlite')

    remove_from_dqlite_cluster(address)

    clear_flag('endpoint.ha.departed')
    set_flag('gateway.changed')


@when('endpoint.ha.changed')
def endpoint_ha_changed(_):
    # Refresh number of dqlite nodes to expect in NRPE check
    clear_flag('nagios.configured')

    clear_flag('endpoint.ha.changed')


@hook('stop')
def stop():
    """Remove the charm main snap when charm is removed"""
    if not is_flag_set('anbox-stream-gateway.installed'):
        return

    remove_anbox_package(DEB_PACKAGE_NAME)


@when('api.available')
def configure_api(api):
    api.configure(port=config('port'))


@when('endpoint.client.joined', 'gateway.started')
@when_not('client.token_created')
def endpoint_client_joined(relation):
    """Generate a random service-name and create a token for it"""
    service_name = "juju_{}_{}_".format(
        local_unit(), remote_unit()).replace('/', '_')
    # Add a random ID to avoid conflicts
    service_name += ''.join(random.choices(
        string.ascii_uppercase + string.digits, k=10))

    try:
        token_bytes = check_output(
            'snap run {} '
            'account create {}'.format(
                GATEWAY, service_name), shell=True, stderr=STDOUT)
    except CalledProcessError as ex:
        output = str(ex.output, 'utf-8')
        if "already exists" in output:
            return
        log('Could not create API token for service {}: {} - {}'.format(
            service_name, ex.returncode, output))
        raise ex
    token = token_bytes.decode('ascii').replace('"', '').replace('\n', '')

    allow_insecure = config('use_insecure_tls')
    if not allow_insecure:
        allow_insecure = False

    location = config("location") or ""
    if len(location) == 0:
        location = "{}:{}".format(get_public_address(), config('port'))

    url = 'https://{}'.format(location)
    private_url = 'https://{}:{}'.format(unit_private_ip(), config('port'))
    relation.send_information(
        token=token,
        url=url,
        private_url=private_url,
        allow_insecure=allow_insecure,
    )

    set_flag('client.token_created')


@when('endpoint.client.departed', 'anbox-stream-gateway.installed')
def endpoint_client_departed(_):
    clear_flag('client.token_created')
    clear_flag('endpoint.client.departed')


@when('config.changed', 'anbox-stream-gateway.installed')
def config_changed():
    configure()


@when('prometheus.available')
def prometheus_available():
    # prometheus-manual means TLS and basic auth are configured, which this
    # interface doesn't support
    if is_flag_set('endpoint.prometheus-manual.joined'):
        if relation_ids():
            relation_set(relation_id(), {})
            return

    if relation_id():
        relation_set(
            relation_id(), {
                'port': config('prometheus_port'),
                'metrics_path': config("prometheus_metrics_path")
            })


@when('endpoint.prometheus-manual.changed', 'gateway.ca_set')
def endpoint_prometheus_manual_joined():
    prometheus = endpoint_from_flag('endpoint.prometheus-manual.changed')
    credentials = prometheus_basic_auth()
    if credentials is None:
        status_set('maintenance', 'Waiting for leader to setup Prometheus')
        return

    with open(TLS_CA_CERT) as ca:
        ca_cert = ca.read()
    if ca_cert is None:
        status_set('maintenance', 'CA Certificate not yet available')
        return

    configure()
    addr = unit_private_ip()
    port = config('prometheus_port')
    prometheus.register_job(
        # A unique ID will be appended to job_name
        job_name='anbox-stream-gateway',
        ca_cert=ca_cert,
        job_data={
            'static_configs': [{
                'targets': ["{}:{}".format(addr, port)]
            }],
            'scheme': 'https',
            'metrics_path': config("prometheus_metrics_path"),
            'tls_config': {'ca_file': '__ca_file__'},
            'basic_auth': {
                'username': credentials['username'],
                'password': credentials['password']
            },
        })
    clear_flag('endpoint.prometheus-manual.changed')


@when('endpoint.prometheus.changed')
def endpoint_prometheus_changed():
    configure()
    clear_flag('endpoint.prometheus.changed')


@when('endpoint.grafana.joined')
@when_any('prometheus.available', 'endpoint.prometheus-manual.available')
@when_not('grafana_dashboard.registered')
def register_grafana_dashboards():
    grafana = endpoint_from_flag('endpoint.grafana.joined')
    with open('templates/grafana_dashboard.json') as dashboard_file:
        json_dashboard = json.loads(dashboard_file.read())
        grafana.register_dashboard(name='anbox-streaming-stack-dashboard',
                                   dashboard=json_dashboard)
    set_flag('grafana_dashboard.registered')


@when('endpoint.grafana.departed')
def grafana_departed():
    clear_flag('grafana_dashboard.registered')
    clear_flag('endpoint.grafana.departed')


@when('certificates.available', 'anbox-stream-gateway.installed')
def prepare_tls_certificates(tls):
    if not os.path.exists(TLS_DIR):
        os.mkdir(TLS_DIR, mode=0o700)

    # Private IP certificate for internal connections
    subject_alt_names = set()
    subject_alt_names.add(socket.gethostname())
    subject_alt_names.add(unit_private_ip())
    subject_alt_names = list(subject_alt_names)
    tls.request_server_cert(unit_private_ip(), subject_alt_names)

    # Public IP certificate for public facing connections
    subject_alt_names = set()
    subject_alt_names.add(socket.gethostname())
    subject_alt_names.add(unit_public_ip())
    subject_alt_names = list(subject_alt_names)
    tls.request_server_cert(unit_public_ip(), subject_alt_names)


def remove_if_exists(path):
    if not os.path.exists(path):
        return
    os.remove(path)


# NOTE: It is important that we wait here for the HA endpoint to
# be removed before we also remove the certificates. Otherwise
# the other dqlite peers will fail to talk to the still existing
# node (but now without TLS) and error out:
#
# Dqlite: ERROR: proxy: first: remote -> local: tls: first record
#   does not look like a TLS handshake
#
# To avoid this we wait until the HA relation is fully departed
# and only then remove all certificates. As the *.departed flag
# is sticky until we clear it the departed handler below will
# be considered automatically when the endpoint.ha.joined flag
# is cleared
@when('endpoint.certificates.departed')
def certificates_relation_departed():
    clear_flag('gateway.certificates_set')
    clear_flag('gateway.ca_set')

    # Remove all certificates we have consumed from the relation
    remove_if_exists(TLS_CA_CERT)
    remove_if_exists(TLS_PUBLIC_CERT)
    remove_if_exists(TLS_PUBLIC_KEY)
    remove_if_exists(TLS_PRIVATE_CERT)
    remove_if_exists(TLS_PRIVATE_KEY)

    clear_flag('endpoint.certificates.departed')


@when_any('certificates.ca.changed', 'certificates.ca.available')
@when('anbox-stream-gateway.installed')
@when_not('endpoint.certificates.departed')
def install_root_ca_cert():
    # When re-joining the relation, .changed is not triggered and we
    # have to rely on .available with a custom flag to avoid writing
    # certificates many times
    if is_flag_set('gateway.ca_set'):
        return
    ca = endpoint_from_flag('certificates.ca.available')
    status_set('maintenance', 'Configuring CA')
    ca_cert = ca.get_ca()
    log('Certificates: setting up CA certificate', 'INFO')
    if ca_cert is None:
        status_set('blocked', 'Failed to communicate with CA')
        return

    ensure_service_dir()

    fmode = os.O_CREAT | os.O_WRONLY
    with open(os.open(TLS_CA_CERT, fmode, 0o600), "w+") as f:
        f.write(ca.get_ca())

    set_flag('gateway.ca_set')
    configure()
    clear_flag('certificates.ca.changed')


@when_any('certificates.certs.changed', 'certificates.certs.available')
@when('anbox-stream-gateway.installed')
def update_certs():
    if is_flag_set('gateway.certificates_set'):
        return
    ca = endpoint_from_flag('certificates.certs.available')
    if not ca.server_certs:
        return

    status_set('maintenance', 'Configuring CA')
    log('Certificates: updating certificates', 'INFO')

    place_certs(ca)

    set_flag('gateway.certificates_set')
    set_flag('gateway.changed')
    clear_flag('certificates.certs.changed')


def place_certs(ca):
    """
    place_certs gets all certificates from the CA relation and puts them
    in their appropriate location
    """
    # For each unit, we can only have two certificates as we use the public
    # and private IP as CN and the relation ignores duplicates
    private_ip_cert = None
    public_ip_cert = None

    for cert in ca.server_certs:
        if cert.common_name == unit_public_ip():
            public_ip_cert = cert
        elif cert.common_name == unit_private_ip():
            private_ip_cert = cert
        else:
            log('Certificates: invalid CN ' + cert.common_name, 'ERROR')
            status_set("blocked", "Error setting up certificates")
            return

        # The tls-certificates relation ignore multiple cert requests with
        # the same CN, so when private_ip == public_ip, we use the same cert
        if unit_private_ip() == unit_public_ip():
            log('Certificates: Private IP is same as public IP, using same certificate', 'WARN')
            private_ip_cert = public_ip_cert

    if private_ip_cert is None or public_ip_cert is None:
        log('Certificates: Empty certificates', 'ERROR')
        status_set("blocked", "Error setting up certificates")
        return

    fmode = os.O_CREAT | os.O_WRONLY
    # Private IP certificate
    with open(os.open(TLS_PRIVATE_KEY, fmode, 0o600), "w+") as f:
        f.write(private_ip_cert.key)
    with open(os.open(TLS_PRIVATE_CERT, fmode, 0o600), "w+") as f:
        f.write(private_ip_cert.cert)

    # Public IP certificate
    with open(os.open(TLS_PUBLIC_KEY, fmode, 0o600), "w+") as f:
        f.write(public_ip_cert.key)
    with open(os.open(TLS_PUBLIC_CERT, fmode, 0o600), "w+") as f:
        f.write(public_ip_cert.cert)

    log('Certificates: Updated certificates', 'INFO')


@hook('leader-elected')
def leader_elected():
    if is_flag_set('anbox-stream-gateway.installed'):
        configure()


@hook('leader-settings-changed')
def leader_settings_changed():
    if is_flag_set('anbox-stream-gateway.installed'):
        configure()


@hook('upgrade-charm')
def upgrade_charm():
    if not is_flag_set('anbox-stream-gateway.installed'):
        return

    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)

    ensure_service_dir()
    set_charm_version()
    configure()

    # If we have an NRPE relation, force run of that as well
    if relation_ids('nrpe-exteral-master'):
        log('Triggering nagios configuration', level='INFO')
        clear_flag('nagios.configured')

    # Refresh grafana dashboard
    if relation_ids('grafana'):
        log('Triggering grafana configuration', level='INFO')
        clear_flag('grafana_dashboard.registered')


def get_lb_address_if_available():
    api = endpoint_from_flag('endpoint.api.joined')
    if api:
        for name in api.all_joined_units.keys():
            if not name.startswith("anbox-stream-gateway-lb"):
                continue
            address = api.all_joined_units[name].received['public-address']
            return address
    return ""


@when('ua.ready')
@when_not('nats.configured', 'anbox-stream-gateway.installed')
def nats_unavailable():
    status_set('blocked', 'Waiting for NATs')


@when('ua.ready')
@when_not('gateway.ca_set', 'gateway.certificates_set', 'anbox-stream-gateway.installed')
def certs_unavailable():
    status_set('blocked', 'Waiting for CA')


@when('nats.configured',
      'gateway.ca_set',
      'gateway.certificates_set',
      'anbox-stream-gateway.installed')
@when_not('gateway.primed')
def gateway_primed():
    set_flag('gateway.primed')
    set_flag('gateway.changed')


@when_not_all('nats.configured',
              'gateway.ca_set',
              'gateway.certificates_set',
              'anbox-stream-gateway.installed')
def gateway_not_primed():
    clear_flag('gateway.primed')


@when('gateway.started')
@when_not('gateway.primed')
def clean_before_stop():
    """
    Essential relations have departed while the gateway was running
    Perform cleanup before final shutdown
    """
    if is_flag_set('endpoint.ha.joined'):
        address = kv().get('dqlite-address')
        if len(address) == 0:
            raise Exception('Missing dqlite address in order to remove node')
        remove_from_dqlite_cluster(address)
        clear_flag('dqlite.registered')

    if not os.path.exists(GATEWAY_CONFIG_PATH):
        raise Exception('Gateway was running without configuration file')

    os.remove(GATEWAY_CONFIG_PATH)
    service_stop(SERVICE)
    clear_flag('gateway.started')


@when('gateway.started', 'endpoint.ha.joined')
@when_not('dqlite.registered')
def broadcast_dqlite_to_cluster():
    ha = endpoint_from_flag('endpoint.ha.joined')
    port = get_local_dqlite_port()
    ha.register('{}:{}'.format(unit_private_ip(), port))
    set_flag('dqlite.registered')


@when('gateway.changed', 'gateway.primed')
def configure():
    # If we have only CA or certificates set we have to wait until we have both
    # or nothing to avoid an unneeded gateway restart
    if (not is_flag_set('gateway.ca_set') or
            not is_flag_set('gateway.certificates_set')):
        status_set('blocked', 'Waiting for CA and certificates')
        return False

    nats = kv().get('nats_config')
    if nats is None:
        # In theory this shouldn't happen, but there may be some Juju delay
        # between Nats removal and gateway_not_primed being called
        return

    status_set('maintenance', 'Configuring the stream gateway')

    port = config('port')
    listen_address = unit_private_ip()
    public_address = unit_public_ip()
    if listen_address == public_address:
        listen_address = get_public_address()
        public_address = listen_address

    store_driver = 'dqlite'
    dqlite_conf = get_dqlite_config(listen_address)
    if dqlite_conf is None:
        status_set('blocked', 'Waiting for cluster to be established')
        return False
    store_options = dump(dqlite_conf, Dumper=Dumper)
    kv().set('dqlite-address', dqlite_conf['listen-address'])

    dev_ui = config('enable_dev_ui')
    if not dev_ui:
        dev_ui = False

    use_insecure_tls = config('use_insecure_tls')
    if not use_insecure_tls:
        use_insecure_tls = False

    keep_containers = config('keep_containers')
    if not keep_containers:
        keep_containers = False

    # We have to store the auth secret separately to avoid changing
    # it when we render the configuration multiple times
    if is_leader():
        if not leader_get('auth_secret'):
            secret_file_exists = False
            if not os.path.exists(AUTH_SECRET_PATH):
                auth_secret = generate_random_string(64)
            else:
                with open(AUTH_SECRET_PATH, "r") as f:
                    auth_secret = f.read()
                    secret_file_exists = True

            auth_info = {'auth_secret': "{}".format(auth_secret)}
            leader_set(auth_info)

            if secret_file_exists:
                os.remove(AUTH_SECRET_PATH)
    else:
        if leader_get('auth_secret') is None:
            status_set('maintenance', 'Waiting for leader to setup secret key')
            return False

    # In case that we have no location but a load balancer (all default
    # deployments) we have to take special care of setting the right location
    # for the gateway configuration so that all out handed API URLs using the
    # correct address
    location = config('location') or get_lb_address_if_available()

    # Metrics setup
    prometheus_listen_address = prometheus_username = prometheus_password = ""
    prom_tls_key = prom_tls_cert = ""
    enable_prometheus = is_flag_set('prometheus.available') or \
        is_flag_set('endpoint.prometheus-manual.joined')
    if enable_prometheus:
        close_old_and_open_new_ports('prometheus_port')
        prom_ip = unit_private_ip()
        prometheus_listen_address = "{}:{}".format(prom_ip, config('prometheus_port'))

        # The old 'prometheus' interface does not support TLS or basic auth
        if is_flag_set('endpoint.prometheus-manual.joined'):
            basic_auth = prometheus_basic_auth()
            if basic_auth is None:
                status_set('maintenance', 'Waiting for leader to setup Prometheus')
                return
            prometheus_username = basic_auth['username']
            prometheus_password = basic_auth['password']
            prom_tls_cert = TLS_PRIVATE_CERT
            prom_tls_key = TLS_PRIVATE_KEY

    templating.render(
        source='config.yaml',
        target=GATEWAY_CONFIG_PATH,
        context={
            'log_level': config('log_level'),
            'location': "{}".format(location),
            'listen_address': "{}:{}".format(listen_address, port),
            'public_address': "{}:{}".format(public_address, port),
            'req_per_second': config('max_http_requests_per_second') or 100,
            'auth_secret': leader_get('auth_secret'),
            'dev_ui': dev_ui,
            'use_insecure_tls': use_insecure_tls,
            'keep_containers': keep_containers,
            'nats_urls': nats['endpoints'],
            'nats_tls_ca_path': nats['tls']['ca'] if nats.get('tls') else [],
            'nats_tls_ca': nats.get('tls') is not None,
            'store_driver': store_driver,
            'store_options': store_options,
            'tls_key_path': TLS_PUBLIC_KEY,
            'tls_cert_path': TLS_PUBLIC_CERT,
            'preferred_region': config("preferred_region") or "",
            'enable_pprof': config('enable_pprof') or False,
            'enable_prometheus': enable_prometheus,
            'prometheus_listen_address': prometheus_listen_address,
            'prometheus_tls_key': prom_tls_key,
            'prometheus_tls_cert': prom_tls_cert,
            'prometheus_basic_auth_username': prometheus_username,
            'prometheus_basic_auth_password': prometheus_password,
        },
        perms=0o600)

    close_old_and_open_new_ports('port')

    # Now that we have a configuration file in place we can enable the service
    service('enable', SERVICE)

    log('Everything configured, starting gateway')

    # We catch potential errors to not block the charm from running other
    # hooks which could potentially help the gateway run again
    try:
        if not is_snap_connected('anbox-stream-gateway', 'daemon-notify'):
            setup_gateway_as_notify_service()

        service_restart(SERVICE)
        set_flag('gateway.started')
    except CalledProcessError:
        status_set('blocked', 'Gateway failed to start. Check logs.')

    status_set('active', '')
    set_flag('gateway.started')
    clear_flag('gateway.changed')
    return True


def get_local_dqlite_port():
    # we generate the dqlite port from a base and its unique ID.
    # ex: with an ID of 8, port will be 7000 + 8 = 7008
    unit_id = int(local_unit().split('/')[1])
    port = DQLITE_PORT_BASE + unit_id
    return port


def get_dqlite_config(listen_address):
    log('Dqlite: configuring {}'.format(local_unit()), 'INFO')
    if not os.path.exists(SNAP_DQLITE_PATH):
        os.mkdir(SNAP_DQLITE_PATH, mode=0o700)
    if not os.path.exists(DQLITE_DATA_DIR):
        os.mkdir(DQLITE_DATA_DIR, mode=0o700)

    port = get_local_dqlite_port()

    # Because a port is tied to a unit ID, there is no need to close
    # as a new unit ID would mean a new machine
    open_port(port)

    store_options = {
        'listen-address': '{}:{}'.format(listen_address, port),
        'data-dir': DQLITE_DATA_DIR,
    }

    log('Dqlite: configuring with TLS', level='INFO')
    store_options['tls'] = {
        'allow-insecure': False,
        'cert': TLS_PRIVATE_CERT,
        'key': TLS_PRIVATE_KEY,
        'ca-cert': TLS_CA_CERT
    }

    # If we're not the leader and there is no HA relation yet there is
    # no valid configuration we can give at this time
    ha = endpoint_from_flag('endpoint.ha.joined')
    if not is_leader() and not ha:
        return None

    # If we're the leader and have no peers yet the array will remain
    # empty. If we have an active HA relation, peers are taken and used
    peers = []
    if ha:
        peers = [{'address': p} for p in ha.peers()]
        log("Dqlite: Peer relation is joined and we have {} peers".format(len(peers)))

    # If we're not the leader and we have an empty peers list we're
    # not alone and have to wait for all peer information to be published
    if not is_leader() and len(peers) == 0:
        return None

    store_options['peers'] = peers

    return store_options


def remove_from_dqlite_cluster(address):
    """"
    Explicitly tell the dqlite cluster to not consider this node as down
    but rather as definitely removed
    """
    log("Removing {} from dqlite cluster".format(address), 'INFO')
    try:
        check_output([
            GATEWAY, 'cluster', 'remove', address, '--force'])
    except CalledProcessError as ex:
        log('Failed to remove {} from dqlite cluster: {} - {}'.format(
            address, ex.returncode, str(ex.output, 'utf-8')), "ERROR")
        raise ex


# check_snap_connected returns true if the given snap name is
# connected to the given interface
def is_snap_connected(snap, interface):
    session = requests_unixsocket.Session()
    resp = session.get('http+unix://%2Frun%2Fsnapd.socket/v2/connections')
    connections = resp.json()['result']['established']
    for connection in connections:
        if connection['interface'] == interface \
                and connection['plug']['snap'] == snap:
            return True
    return False


def prometheus_basic_auth():
    """
    Get access credentials for Prometheus.

    If a configuration item for the username or the password is set
    it takes precedence over the rest. If nothing exists yet, a new set of
    credentials is created and stored on the leader relation.
    :return: dictionary with keys 'username' and 'password' or None if
            something went wrong or if credentials are not yet configured
    """
    current_username = leader_get('prometheus_username')
    current_password = leader_get('prometheus_password')

    if not is_leader():
        return {'username': current_username, 'password': current_password}

    cfg_username = config('prometheus_basic_auth_username')
    cfg_password = config('prometheus_basic_auth_password')

    generated_password = \
        leader_get('prometheus_generated_password') or generate_random_string(16)

    username = cfg_username or 'prometheusadmin'
    password = cfg_password or generated_password

    if current_username != username or current_password != password:
        leader_set(
            prometheus_username=username,
            prometheus_password=password,
            prometheus_generated_password=generated_password
        )

    return {'username': username, 'password': password}


def setup_gateway_as_notify_service():
    check_call(['snap', 'connect', 'anbox-stream-gateway:daemon-notify',
                'system:daemon-notify'])


@when_any('config.changed.nagios_context',
          'config.changed.nagios_servicegroups')
def nagios_config_changed():
    if relation_ids('nrpe-exteral-master'):
        clear_flag('nagios.configured')


@when('nrpe-external-master.available')
@when_not('nagios.configured')
def update_nrpe_config():
    hostname = nrpe.get_nagios_hostname()
    add_user_to_group(NRPE_USER, GATEWAY_GROUP)

    nrpe_setup = nrpe.NRPE(hostname=hostname)

    src = os.path.join(os.environ['CHARM_DIR'], 'files', 'nrpe',
                       'check_anbox_stream_gateway.sh')
    # Files in the charm directory have no guarantee to stay there as per
    # Juju docs, so we move the script in a more stable place
    shutil.copy2(src, '/usr/lib/nagios/plugins')
    cmd = ['/usr/lib/nagios/plugins/check_anbox_stream_gateway.sh']
    ha = endpoint_from_flag('endpoint.ha.joined')
    if ha:
        # all_joined_units only returns OTHER units. We add +1 to get the
        # total number of nodes
        expected_nodes = len(ha.all_joined_units) + 1
        cmd.extend(['--expected-db-nodes={}'.format(str(expected_nodes))])

    nrpe_setup.add_check(
        shortname='anbox-stream-gateway_status',
        description='Check Anbox Stream Gateway service status',
        check_cmd=" ".join(cmd),
    )

    nrpe_setup.write()
    set_flag('nagios.configured')


def deploy_drop_in():
    """Deploys a drop in for the Stream Gateway systemd service to chown
    unix socket to 'anbox-stream-gateway' group after any restart"""
    if not os.path.exists(SERVICE_DROP_IN_FOLDER):
        os.mkdir(SERVICE_DROP_IN_FOLDER, mode=755)

    templating.render(
        source=SERVICE_DROP_IN_NAME,
        target=SERVICE_DROP_IN_PATH,
        context={},
        perms=0o644)

    reload_systemd()
